SystemOrganization addCategory: #'XTream-Parallel'!

Object subclass: #XtreamFIFO
	instanceVariableNames: 'storage readPosition writePosition size readReady writeReady closed endOfStreamAction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Parallel'!

!XtreamFIFO commentStamp: 'nice 12/2/2009 15:33' prior: 0!
A XtreamFIFO is circular queue for synchronizing read and writes

Instance Variables:
	storage	<Collection> where to store the objects
	readPosition	<Integer> between 0 and storage size - 1 = index after which to read the next object
	readReady	<Semaphore> to be waited at next read if queue is empty
	size	<Integer> the size of queue (number of objects to read)
	writePosition	<Integer> between 0 and storage size - 1 = index after which to write the next object
	writeReady	<Semaphore> to be waited at next read if queue is full
	closed	<Boolean> indicating that no further write will be attempted
	endOfStreamAction <Valuable> to be evaluated when an attempt to read an empty closed fifo will occur.
!

!XtreamFIFO methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 15:30'!
close
	closed := true! !

!XtreamFIFO methodsFor: 'acessing' stamp: 'nice 12/2/2009 16:57'!
contentsSpecies
	^storage species! !

!XtreamFIFO methodsFor: 'initialize-release' stamp: 'nice 12/3/2009 09:53'!
endOfStreamAction: aValuable
	| old |
	old := endOfStreamAction.
	endOfStreamAction := aValuable.
	^old! !

!XtreamFIFO methodsFor: 'testing' stamp: 'nice 12/2/2009 19:26'!
isEmpty
	^size = 0! !

!XtreamFIFO methodsFor: 'testing' stamp: 'nice 12/2/2009 19:26'!
isFull
	^size = storage size! !

!XtreamFIFO methodsFor: 'reading' stamp: 'nice 12/2/2009 19:59'!
next
	| index result |
	self isEmpty ifTrue: [
		closed
			ifTrue: [
				Processor yield. "Give a chance of running to pending write process"
				self isEmpty ifTrue: [^endOfStreamAction value]]
			ifFalse: [self waitReadReady]].
	index := readPosition + 1.
	readPosition := index \\ storage size.
	result := storage at: index.
	size := size - 1.
	writeReady signal.
	^result! !

!XtreamFIFO methodsFor: 'reading' stamp: 'nice 12/2/2009 19:59'!
next: n into: aCollection startingAt: startIndex
	| index nread |
	n = 0 ifTrue: [^self].
	self isEmpty ifTrue: [
		closed
			ifTrue: [
				Processor yield. "Give a chance to run to pending write process"
				self isEmpty ifTrue: [^endOfStreamAction value]]
			ifFalse: [self waitReadReady]].
	index := readPosition + 1.
	nread := (size min: n) min: storage size - readPosition.
	aCollection replaceFrom: startIndex to: -1 + nread + startIndex with: storage startingAt: index.
	size := size - nread.
	readPosition := readPosition + nread \\ storage size.
	writeReady signal.
	nread = n ifTrue: [^self].
	^self next: n - nread into: aCollection startingAt: startIndex + nread! !

!XtreamFIFO methodsFor: 'writing' stamp: 'nice 12/2/2009 19:59'!
next: n putAll: aCollection startingAt: startIndex
	| index  nwrite |
	n = 0 ifTrue: [^self].
	self isFull ifTrue: [self waitWriteReady].
	index := writePosition + 1.
	nwrite := (storage size - size min: n) min: (storage size - writePosition).
	storage replaceFrom: index to: -1 + nwrite + index with: aCollection startingAt: startIndex.
	size := size + nwrite.
	writePosition := writePosition + nwrite \\ storage size.
	readReady signal.
	nwrite = n ifTrue: [^self].
	^self next: n - nwrite putAll: aCollection startingAt: startIndex + nwrite! !

!XtreamFIFO methodsFor: 'writing' stamp: 'nice 12/2/2009 20:00'!
nextPut: anObject
	| index  |
	self isFull ifTrue: [self waitWriteReady].
	index := writePosition + 1.
	writePosition := index \\ storage size.
	storage at: index put: anObject.
	size := size + 1.
	readReady signal.
	^anObject! !

!XtreamFIFO methodsFor: 'reading' stamp: 'nice 12/2/2009 19:26'!
readInto: aCollection startingAt: startIndex count: n
	"Non blocking read.
	Answer the number of elements read"
	
	| index nread |
	n = 0 ifTrue: [^0].
	self isEmpty ifTrue: [
		closed
			ifTrue: [
				Processor yield. "Give a chance to run to pending write process"
				self isEmpty ifTrue: [endOfStreamAction value. ^0]]
			ifFalse: [^0]].
	index := readPosition + 1.
	nread := (size min: n) min: storage size - readPosition.
	aCollection replaceFrom: startIndex to: -1 + nread + startIndex with: storage startingAt: index.
	size := size - nread.
	readPosition := readPosition + nread \\ storage size.
	writeReady signal.
	nread = n ifTrue: [^n].
	^nread + (self readInto: aCollection startingAt: startIndex + nread count: n - nread)! !

!XtreamFIFO methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 16:27'!
reopen
	closed := false! !

!XtreamFIFO methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 15:28'!
storage: aCollection
	readReady := Semaphore new.
	writeReady := Semaphore new.
	readPosition := writePosition := size := 0.
	storage := aCollection.
	closed := false! !

!XtreamFIFO methodsFor: 'reading' stamp: 'nice 12/2/2009 19:58'!
waitReadReady
	"It might be necessary to loop due to excess signals"
	
	[self isEmpty] whileTrue: [readReady wait]! !

!XtreamFIFO methodsFor: 'writing' stamp: 'nice 12/2/2009 19:59'!
waitWriteReady
	"It might be necessary to loop due to excess signals"
	
	[self isFull] whileTrue: [writeReady wait]! !

!XtreamFIFO methodsFor: 'writing' stamp: 'nice 12/2/2009 19:27'!
writeFrom: aCollection startingAt: startIndex count: n
	"Non blocking write.
	answer the number of elements written"
	
	| index  nwrite |
	n = 0 ifTrue: [^0].
	self isFull ifTrue: [^0].
	index := writePosition + 1.
	nwrite := (storage size - size min: n) min: (storage size - writePosition).
	storage replaceFrom: index to: -1 + nwrite + index with: aCollection startingAt: startIndex.
	size := size + nwrite.
	writePosition := writePosition + nwrite \\ storage size.
	readReady signal.
	nwrite = n ifTrue: [^n].
	^nwrite + (self writeFrom: aCollection startingAt: startIndex + nwrite count: n - nwrite)! !

!Xtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 08:14'!
collecting: aBlock
	^self transforming: [:input :output |
		output nextPut: (aBlock value: input next)]! !

!Xtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 08:14'!
collecting: aBlock as: collectionSpecies
	^self transforming: [:input :output |
		output nextPut: (aBlock value: input next)] as: collectionSpecies! !

!Xtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 08:14'!
rejecting: aBlock
	^self transforming: [:input :output |
		| value |
		[value := input next.
		aBlock value: value] whileTrue.
		output nextPut: value]! !

!Xtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 08:18'!
selecting: aBlock
	^self transforming: [:input :output |
		| value |
		[value := input next.
		aBlock value: value] whileFalse.
		output nextPut: value]! !

!Xtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 08:19'!
transforming: aBlock
	^self transforming: aBlock as: self contentsSpecies! !

!Xtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 09:24'!
transforming: aBlock as: collectionSpecies
	self subclassResponsibility! !

!ReadXtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 08:14'!
transforming: aBlock as: collectionSpecies
	| fifo |
	fifo := XtreamFIFO new storage: (collectionSpecies ofSize: 1024).
	^TransformReadXtream new
		source: self;
		block: aBlock;
		fifo: fifo;
		endOfStreamAction: endOfStreamAction;
		run! !

ReadXtream subclass: #TransformReadXtream
	instanceVariableNames: 'block atEnd fifo'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Parallel'!

!TransformReadXtream commentStamp: 'nice 12/2/2009 04:10' prior: 0!
A TransformReadXtream is a class taking objects in a source stream and transforming.
It is a very interesting objects for piping stream processing.

Instance Variables:
	atEnd		<Boolean> an indicator that source is exhausted
	block		<BlockClosure> taking two arguments: the input source stream and the output fifo
	fifo		<FIFO> a queue for synchronizing reads and writes

Every read to this stream will be forwarded to the fifo.
The fifo is fed by the soure stream via the block.
The block does not need to put as many objects on the fifo as it reads form the source.
This is why some kind of buffering and synchronisation are needed.!

!TransformReadXtream methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 01:01'!
block: aBlock
	"Set the block to be evaluated after reading the source stream"
	block := aBlock! !

!TransformReadXtream methodsFor: 'file API' stamp: 'nice 12/2/2009 19:50'!
close
	atEnd := true.
	source close.
	fifo close! !

!TransformReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:33'!
closed
	^source closed! !

!TransformReadXtream methodsFor: 'accessing' stamp: 'nice 12/2/2009 19:23'!
contents
	| res |
	res := (self contentsSpecies ofSize: 100) writeStream.
	fifo endOfStreamAction: [^res contents].
	[true] whileTrue: [res nextPut: self next]! !

!TransformReadXtream methodsFor: 'accessing' stamp: 'nice 12/2/2009 16:57'!
contentsSpecies
	^fifo contentsSpecies! !

!TransformReadXtream methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 03:43'!
fifo: aFifo
	fifo := aFifo! !

!TransformReadXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:55'!
name
	^source name! !

!TransformReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 03:44'!
next
	^fifo next! !

!TransformReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 03:49'!
next: anInteger into: aCollection startingAt: startIndex
	^fifo next: anInteger into: aCollection startingAt: startIndex! !

!TransformReadXtream methodsFor: 'reading' stamp: 'nice 12/2/2009 03:50'!
readInto: aCollection startingAt: startIndex count: anInteger
	^fifo readInto: aCollection startingAt: startIndex count: anInteger! !

!TransformReadXtream methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 19:49'!
run
	atEnd := false.
	source endOfStreamAction: [EndOfStream signal].
	fifo endOfStreamAction: endOfStreamAction.
	fifo reopen.
	[
	[[atEnd] whileTrue: [block value: source value: fifo]] on: EndOfStream do: [:exc | exc return: nil].
	] fork! !

TestCase subclass: #XTreamParallelTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Parallel'!

!XTreamParallelTest methodsFor: 'testing' stamp: 'nice 12/3/2009 20:32'!
testFIFO
	"XtreamTests new testFIFO"
	| fifo interrupt finished loop handler mark |
	fifo := XtreamFIFO new storage: (ByteArray new: 3).
	fifo next: 3 putAll: #[1 2 3] startingAt: 1.
	self assert: fifo isFull.
	self deny: fifo isEmpty.
	self assert: (fifo writeFrom: #[1 2 3] startingAt: 1 count: 3) = 0.
	self assert: fifo isFull.
	self assert: fifo next = 1.
	self deny: fifo isFull.
	self deny: fifo isEmpty.
	self assert: fifo next = 2.
	self deny: fifo isFull.
	self deny: fifo isEmpty.
	self assert: fifo next = 3.
	self deny: fifo isFull.
	self assert: fifo isEmpty.
	self assert: (fifo readInto: (ByteArray new: 3) startingAt: 1 count: 3) = 0.
	
	"aBlocking read example"
	finished := mark := Object new.
	interrupt := Semaphore new.
	handler := [(Delay forSeconds: 1) wait. interrupt signal] newProcess.
	loop := [finished := fifo next] newProcess.
	handler resume.
	loop resume.
	interrupt wait.
	loop terminate.
	self assert: finished = mark.
	
	"unblocked read example"
	finished := mark := Object new.
	interrupt := Semaphore new.
	handler := [(Delay forSeconds: 1) wait. interrupt signal] newProcess.
	loop := [finished := fifo next] newProcess.
	handler resume.
	loop resume.
	fifo nextPut: 4.
	Processor yield.
	interrupt wait.
	loop terminate.
	self assert: finished = 4.
	self assert: fifo isEmpty.! !

WriteXtream subclass: #TransformWriteXtream
	instanceVariableNames: 'block fifo atEnd'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'XTream-Parallel'!

!TransformWriteXtream commentStamp: 'nice 12/2/2009 04:09' prior: 0!
A TransformWriteXtream is a class transforming objects before writing on a destination stream
It is a very interesting object for piping stream processing.

Instance Variables:
	atEnd		<Boolean> an indicator that source is exhausted
	block		<BlockClosure> taking two arguments: the input fifo and the output destination
	fifo		<FIFO> a queue for synchronizing reads and writes

Every write to this stream will go to the fifo and feed the block.
The block does not need to put as many objects on the destination as it reads form the fifo.
This is why some kind of buffering and synchronisation are needed.!

!TransformWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 03:42'!
block: aBlock
	"Set the block to be evaluated with input and output streams"
	block := aBlock! !

!TransformWriteXtream methodsFor: 'file API' stamp: 'nice 12/2/2009 19:50'!
close
	atEnd := true.
	destination close.
	fifo close! !

!TransformWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:41'!
closed
	^destination closed! !

!TransformWriteXtream methodsFor: 'accessing' stamp: 'nice 12/2/2009 17:00'!
contents
	fifo close.
	Processor yield.
	fifo reopen.
	^destination contents! !

!TransformWriteXtream methodsFor: 'accessing' stamp: 'nice 12/2/2009 16:56'!
contentsSpecies
	^destination contentsSpecies! !

!TransformWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 03:40'!
fifo: aFifo
	fifo := aFifo! !

!TransformWriteXtream methodsFor: 'file API' stamp: 'nice 12/4/2009 02:55'!
name
	^destination name! !

!TransformWriteXtream methodsFor: 'writing' stamp: 'nice 12/2/2009 03:41'!
next: count putAll: aCollection startingAt: startIndex
	fifo next: count putAll: aCollection startingAt: startIndex! !

!TransformWriteXtream methodsFor: 'writing' stamp: 'nice 3/19/2010 19:06'!
nextPut: anObject
	^fifo nextPut: anObject! !

!TransformWriteXtream methodsFor: 'initialize-release' stamp: 'nice 12/2/2009 19:17'!
run
	atEnd := false.
	fifo endOfStreamAction: [EndOfStream signal].
	fifo reopen.
	[
	[[atEnd] whileFalse: [block value: fifo value: destination]] on: EndOfStream do: [:exc | exc return: nil].
	] fork! !

!WriteXtream methodsFor: '*XTream-Parallel' stamp: 'nice 12/3/2009 08:13'!
transforming: aBlock as: collectionSpecies
	| fifo |
	fifo := XtreamFIFO new storage: (collectionSpecies ofSize: 1024).
	^TransformWriteXtream new
		destination: self;
		block: aBlock;
		fifo: fifo;
		run! !
